#include <imath.h>
#include <stdlib.h>
#include <math.h>

/****************************************************************************/
/****************************************************************************/

/***************************************************/
/*                                                 */
/* Ces routines implmentent l'algorithme          */
/* de dcimation de Mallat. La premire phase      */
/* consiste  calculer les moyennes, la seconde    */
/*  calculer les diffrences. Le tout est O(2n),  */
/* ce qui est quand mme trs efficace.            */
/*                                                 */
/* Aucune contraintes quant  la dimension du vec- */
/* teur. Le DC term est au bout du vecteur (et non */
/*  l'indice zro)                                */
/*                                                 */
/* conserve ~ 20 bits de precision (sur 23), ou    */
/* environ 1e-7 (0.0000001)                        */
/*                                                 */
/***************************************************/

/***************************************************/
/*                                                 */
/* Those two routines implement the Mallat decima- */
/* tion algorithm. First phase computes the means, */
/* the second phase the differences. All is O(2n), */
/* Which is rather efficient.                      */
/*                                                 */
/* keeps ~20 bits of precision (out of 23), or     */
/* a precision of about 1e07                       */
/*                                                 */
/***************************************************/

/*******************************/
void FMT(const float source[], float destination[], int length)
 { 
  float *t;
  int i,d,j;

  t = (float *)malloc(2*length*sizeof(float));

  for (i=0;i<length;i++) t[i]=source[i];

  for (i=0,d=length;i<2*length-2;i+=2,d++) t[d] = t[i]+t[i+1];

  destination[length-1]=t[2*length-2];

  for ( i=2*length-2, j=2*length-3, d=length-2; 
        i>length-1; 
        i--, j-=2, d--) 
    destination[d] = t[j]-t[i]/2.0f;

  free(t);
 }

/*******************************/
void iFMT(const float source[], float destination[], int length)
 {
  float *t; 
  int i,d,p;

  t = (float *)malloc(2*length*sizeof(float));

  t[2*length-2]=source[length-1];

  for ( p=length-2, i=2*length-4 , d=2*length-2; 
       i>=0; i-=2, d--, p--)
   {
    float t0 = t[d]/2;
    float s0 = source[p];
    t[i]   = t0 - s0;
    t[i+1] = t0 + s0;
   }

  for (i=0;i<length;i++) destination[i]=t[i];

  free(t);
 }

/****************************************************************************/
/****************************************************************************/

/*********************************************************/
/*                                                       */
/* Ces routines calculent la transforme rapide de Haar  */
/* normalise.                                           */
/*                                                       */
/* Naturellement, fonctionne avec un vecteur de longueur */
/* quelconque (mais > 3)                                 */
/*                                                       */
/* Les esseuls sont laisss pour compte (puisqu'on ne   */
/* peux pas toujours les pairer avec d'autres coeffi-    */
/* cients) mais laisss dans le vecteur rsultat.        */
/* La prcision conserve est suprieure  6 dcimales,  */
/* qui est ce qui est suppos tre conserv par un float */
/* (bien qu'en pratique, il y en a un peu plus).         */
/*                                                       */
/* conserve ~ 20 bits de precision (sur 23), ou environ  */
/* 1e-7 (0.0000001)                                      */
/*                                                       */
/*********************************************************/

/*********************************************************/
/*                                                       */
/* These routines compute the fast (normalized) Haar     */
/* transform.                                            */
/*                                                       */
/* Works of course with a vector of any lenght > 3.      */
/*                                                       */
/* Lonely coefficients are left as is, since we can't    */
/* pair them correctly, but are put in the result vec-   */
/* tor. The precision is also about ~20 bits (out of 23) */
/*                                                       */
/*********************************************************/

/*******************************/
void FHT(const float source[], float destination[], int length)
 {
  int i,l;

  float *t = (float *)malloc(length*sizeof(float));
  
  for (i=0;i<length;i++) destination[i]=source[i];

  l=length;
  for (i=Log2(length-1);i;i--)
   {
    int j,k;
    int l2 = l/2;

    for(j=0;j<l;j++) t[j]=destination[j];

    for (k=0,j=0;k<l2;k++,j+=2)
     {
      destination[k]    = (t[j]+t[j+1])*0.7071067812f;
      destination[k+l2] = (t[j]-t[j+1])*0.7071067812f;
     }
    
    l /= 2 ;  // for the leftovers
   }

  free(t);
 }

/*******************************/
void iFHT(const float source[], float destination[], int length)
 {
  int i,j;
  int lim[32];
  
  int    k  = Log2(length-1);
  float *t  = (float *)malloc( length * sizeof(float));

  if (( 1 << k) != length) k--; // i.e., length n'est pas une puissance de 2

  for ( j=length, i=k; i; i--, j/=2 ) lim[i]=j;
  lim[0]=1;

  for ( i=0; i<length; i++ ) destination[i]=source[i];

  for(i=0;i<k;i++)
   {
    int d;
    int l=lim[i];
    int l2=lim[i+1];

    for(j=0; j<l2; j++) t[j]=destination[j];

    for( d=0, j=0; j<l; j++, d+=2 )
     {
      destination[d]   =(t[j]+t[j+l])*0.7071067812f;
      destination[d+1] =(t[j]-t[j+l])*0.7071067812f;
     }

   }

  free(t);
 }

